import inspect


KEYS_SEPARATOR = '||'


class DynamoDBQueryException(Exception):
    pass


class FieldDoesNotExists(DynamoDBQueryException):
    pass


class InvalidOperator(DynamoDBQueryException):
    pass


class NoneExecution:

    last_evaluated_key = None


class DynamoDBQuerySetError(Exception):
    pass


class MultipleObjectsReturned(DynamoDBQuerySetError):
    pass


class ObjectDoesNotExist(DynamoDBQuerySetError):
    pass


class DynamoDBQuerySet:

    def __init__(self, model):
        self._model = model
        self._limit = None
        self._last_evaluated_key = None
        self._hash_key = None
        self._index = None
        self._scan_index_forward = True
        self._filter = None
        self._next_key = None
        self._init()

    def _clone(self):
        """
        Return a copy of the current QuerySet. A lightweight alternative
        to deepcopy().
        """
        c = self.__class__(model=self._model)
        c._limit = self._limit
        c._last_evaluated_key = self._last_evaluated_key
        c._hash_key = self._hash_key
        c._index = self._index
        c._scan_index_forward = self._scan_index_forward
        c._filter = self._filter
        if isinstance(self._execution, NoneExecution):
            c._execution = self._execution
            c._results = self._results
        return c

    def __iter__(self):
        self._execute()
        return iter(self._results)

    def _init(self):
        self._execution = None
        self._results = None

    def _process_execution(self):
        query = self._index.query if self._index else self._model.query

        if self._hash_key is None:
            raise DynamoDBQuerySetError('Hash key must be set')

        self._execution = query(
            self._hash_key,
            self._filter,
            limit=self._limit,
            last_evaluated_key=self._last_evaluated_key,
            scan_index_forward=self._scan_index_forward
        )
        self._results = list(self._execution)
        self._next_key = self._execution.last_evaluated_key

    def _execute(self):
        if not self._execution:
            self._process_execution()

    @property
    def next_key(self):
        self._execute()
        return self._next_key

    def set_limit(self, limit):
        obj = self._clone()
        obj._limit = limit
        return obj

    def set_last_evaluated_key(self, last_evaluated_key):
        obj = self._clone()
        obj._last_evaluated_key = last_evaluated_key
        return obj

    def set_index(self, index):
        obj = self._clone()
        obj._index = index
        return obj

    def set_hash_key(self, hash_key):
        obj = self._clone()
        obj._hash_key = hash_key
        return obj

    def none(self):
        obj = self._clone()
        obj._execution = NoneExecution()
        obj._results = []
        return obj

    def set_scan_index_forward(self, value):
        obj = self._clone()
        obj._scan_index_forward = value
        return obj

    def last(self):
        obj = self._clone()
        obj._limit = 1
        obj._scan_index_forward = not obj._scan_index_forward
        obj._execute()
        if obj._results:
            return obj._results[0]
        else:
            return None

    def first(self):
        obj = self._clone()
        obj._limit = 1
        obj._execute()
        if obj._results:
            return obj._results[0]
        else:
            return None

    def exists(self):
        return self.first() is not None

    def _parse_lookup(self, lookup):
        if '__' in lookup:
            return lookup.split('__')
        else:
            return lookup, None

    def _get_field(self, field_name):
        try:
            return getattr(self._model, field_name)
        except AttributeError:
            raise FieldDoesNotExists('Field "{}" does not exist'.format(field_name))

    def _get_filter(self, field, operator, value):
        if operator == 'eq' or operator is None:
            return field == value
        elif operator == 'not':
            return field != value
        elif operator == 'lt':
            return field < value
        elif operator == 'lte':
            return field <= value
        elif operator == 'gt':
            return field > value
        elif operator == 'gte':
            return field >= value
        elif operator == 'between':
            return field.between(*value)
        elif operator == 'in':
            return field.is_in(value)
        elif operator == 'exists' and value:
            return field.exists()
        elif operator == 'exists' and not value:
            return field.does_not_exist()
        elif operator == 'startswith':
            return field.startswith(value)
        elif operator == 'contains':
            return field.contains(value)
        else:
            raise InvalidOperator('Invalid operator "{}"'.format(operator))

    def _pre_filter(self, field, field_name, operator, value):
        pass

    def filter(self, **kwargs):
        assert len(kwargs) == 1, 'Only one filter value is required'

        obj = self._clone()

        lookup, value = list(kwargs.items())[0]

        field_name, operator = obj._parse_lookup(lookup)

        field = obj._get_field(field_name)

        obj._pre_filter(field, field_name, operator, value)
        obj._filter = obj._get_filter(field, operator, value)
        if operator == 'between' and value[0] > value[1]:
            return self.none()
        return obj

    def get(self, **kwargs):
        obj = self
        if kwargs:
            obj = self.filter(**kwargs)
        obj._execute()
        if len(obj._results) == 1:
            return obj._results[0]
        elif len(obj._results) == 0:
            raise ObjectDoesNotExist
        else:
            raise MultipleObjectsReturned

    def count(self):
        if self._execution is not None:
            return len(self._results)
        if self._last_evaluated_key or self._limit:
            self._execute()
            return len(self._results)
        else:
            query = self._index.count if self._index else self._model.count

            if self._hash_key is None:
                raise DynamoDBQuerySetError('Hash key must be set')
            return query(
                self._hash_key,
                self._filter,
            )

    def delete(self):
        with self._model.batch_write() as batch:
            for obj in self:
                batch.delete(obj)

    def as_manager(cls):
        return DynamoDBManager.from_queryset(cls)()
    as_manager.queryset_only = True
    as_manager = classmethod(as_manager)


class BaseDynamoDBManager:

    def __init__(self, index=None):
        super().__init__()
        self.model = None
        self._index = index

    def contribute_to_class(self, model):
        self.model = model

    def get_queryset(self):
        queryset = self._queryset_class(self.model)
        if self._index:
            return queryset.set_index(self._index)
        else:
            return queryset

    @classmethod
    def _get_queryset_methods(cls, queryset_class):
        def create_method(name, method):
            def manager_method(self, *args, **kwargs):
                return getattr(self.get_queryset(), name)(*args, **kwargs)
            manager_method.__name__ = method.__name__
            manager_method.__doc__ = method.__doc__
            return manager_method

        new_methods = {}
        for name, method in inspect.getmembers(queryset_class, predicate=inspect.isfunction):
            # Only copy missing methods.
            if hasattr(cls, name):
                continue
            # Only copy public methods or methods with the attribute `queryset_only=False`.
            queryset_only = getattr(method, 'queryset_only', None)
            if queryset_only or (queryset_only is None and name.startswith('_')):
                continue
            # Copy the method onto the manager.
            new_methods[name] = create_method(name, method)
        return new_methods

    @classmethod
    def from_queryset(cls, queryset_class, index=None, class_name=None):
        if class_name is None:
            class_name = '%sFrom%s' % (cls.__name__, queryset_class.__name__)
        return type(class_name, (cls,), {
            '_queryset_class': queryset_class,
            '_index': index,
            **cls._get_queryset_methods(queryset_class),
        })


class DynamoDBManager(BaseDynamoDBManager.from_queryset(DynamoDBQuerySet)):
    pass



from django.conf import settings

from pynamodb.models import MetaModel, Model

from .connection import TableConnection
from .queryset import DynamoDBManager


dynamodb_model_classes = []


class DynamoMetaModel(MetaModel):

    def __init__(cls, name, bases, attrs):
        if 'objects' not in attrs:
            cls.objects = attrs['objects'] = DynamoDBManager()

        attr_meta = attrs.get('Meta', None)

        abstract = getattr(attr_meta, 'abstract', False)
        isproxy = getattr(attr_meta, 'proxy', False)

        base_meta = getattr(cls, '_meta', None)
        meta = attr_meta or getattr(cls, 'Meta', None)

        # Meta inheritance
        if base_meta:
            for k, v in base_meta.__dict__.items():
                if not (k.startswith('_') or k in {'abstract', 'proxy'} or hasattr(meta, k)):
                    setattr(meta, k, v)

        cls._meta = meta

        super().__init__(name, bases, attrs)

        if meta and not hasattr(cls.Meta, 'billing_mode'):
            meta.billing_mode = settings.PYDJAMODB_DATABASE.get('BILLING_MODE')

        if not abstract and not isproxy:
            dynamodb_model_classes.append(cls)

        for k, v in attrs.items():
            if isinstance(v, DynamoDBManager):
                v.contribute_to_class(cls)


class DynamoModel(Model, metaclass=DynamoMetaModel):

    @classmethod
    def _get_connection(cls):
        if cls._connection is None:
            cls._connection = TableConnection(cls.Meta.table_name)
        return cls._connection

    @classmethod
    def delete_table(cls, wait=False):
        return cls._get_connection().delete_table(wait)

    def __eq__(self, other):
        return repr(self) == repr(other)

    class Meta:
        abstract = True



import time

from django.conf import settings

from pynamodb.connection.table import TableConnection as BaseTableConnection
from pynamodb.exceptions import TableDoesNotExist

from botocore.client import ClientError


class TableConnection(BaseTableConnection):

    def __init__(self, table_name,
                 region=None,
                 host=None,
                 connect_timeout_seconds=None,
                 read_timeout_seconds=None,
                 max_retry_attempts=None,
                 base_backoff_ms=None,
                 max_pool_connections=None,
                 extra_headers=None,
                 aws_access_key_id=None,
                 aws_secret_access_key=None,
                 aws_session_token=None):
        table_name = '{}-{}'.format(settings.PYDJAMODB_DATABASE['TABLE_PREFIX'], table_name)
        region = settings.PYDJAMODB_DATABASE.get('AWS_REGION') if region is None else region
        host = settings.PYDJAMODB_DATABASE.get('HOST') if host is None else host
        aws_access_key_id = (
            settings.PYDJAMODB_DATABASE.get('AWS_ACCESS_KEY_ID') if aws_access_key_id is None
            else aws_access_key_id
        )
        aws_secret_access_key = (
            settings.PYDJAMODB_DATABASE.get('AWS_SECRET_ACCESS_KEY') if aws_secret_access_key is None
            else aws_secret_access_key
        )
        aws_session_token = (
            settings.PYDJAMODB_DATABASE.get('AWS_SESSION_TOKEN') if aws_session_token is None
            else aws_session_token
        )

        super().__init__(
            table_name,
            region=region,
            host=host,
            connect_timeout_seconds=connect_timeout_seconds,
            read_timeout_seconds=read_timeout_seconds,
            max_retry_attempts=max_retry_attempts,
            base_backoff_ms=base_backoff_ms,
            max_pool_connections=max_pool_connections,
            extra_headers=extra_headers,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            aws_session_token=aws_session_token
        )

    def create_table(self,
                     attribute_definitions=None,
                     key_schema=None,
                     read_capacity_units=None,
                     write_capacity_units=None,
                     global_secondary_indexes=None,
                     local_secondary_indexes=None,
                     stream_specification=None,
                     billing_mode=None,
                     tags=None,
                     wait=False,
                     set_point_in_time_recovery=None):

        billing_mode = settings.PYDJAMODB_DATABASE.get('BILLING_MODE') if billing_mode is None else billing_mode
        set_point_in_time_recovery = (
            settings.PYDJAMODB_DATABASE.get('POINT_IN_TIME_RECOVERY', False) if set_point_in_time_recovery is None
            else set_point_in_time_recovery
        )
        if not stream_specification:
            stream_specification = settings.PYDJAMODB_DATABASE.get('STREAM_SPECIFICATION')

        if tags is None and 'TAGS' in settings.PYDJAMODB_DATABASE:
            tags = {
                k: v.format(table_name=self.table_name) for k, v in settings.PYDJAMODB_DATABASE['TAGS'].items()
            }
        result = self.connection.create_table(
            self.table_name,
            attribute_definitions=attribute_definitions,
            key_schema=key_schema,
            read_capacity_units=read_capacity_units,
            write_capacity_units=write_capacity_units,
            global_secondary_indexes=global_secondary_indexes,
            local_secondary_indexes=local_secondary_indexes,
            stream_specification=stream_specification,
            billing_mode=billing_mode,
            tags=tags
        )

        if wait or set_point_in_time_recovery:
            self.connection.client.get_waiter('table_exists').wait(
                TableName=self.table_name,
                WaiterConfig={
                    'Delay': 10,
                    'MaxAttempts': self.connection._max_retry_attempts_exception
                }
            )

        if set_point_in_time_recovery:
            self.set_point_in_time_recovery(enabled=True)

        return result

    def delete_table(self, wait=False):
        """
        Performs the DeleteTable operation and returns the result
        """
        result = self.connection.delete_table(self.table_name)
        if wait:
            self.connection.client.get_waiter('table_not_exists').wait(
                TableName=self.table_name,
                WaiterConfig={
                    'Delay': 10,
                    'MaxAttempts': self.connection._max_retry_attempts_exception
                }
            )
        return result

    def exists_table(self):
        try:
            self.describe_table()
            return True
        except TableDoesNotExist:
            return False

    def set_point_in_time_recovery(self, enabled=True):
        for i in range(0, self.connection._max_retry_attempts_exception + 1):
            try:
                self.connection.client.update_continuous_backups(
                    TableName=self.table_name,
                    PointInTimeRecoverySpecification={
                        'PointInTimeRecoveryEnabled': enabled
                    }
                )
                break
            except ClientError:
                if i == self.connection._max_retry_attempts_exception:
                    raise
                time.sleep(10)


class TestTableConnection:

    def __init__(self, wrapped_connection, prefix=None):
        self._wrapped_connection = wrapped_connection
        self._is_test_clean_required = False
        if prefix:
            self._wrapped_connection.table_name = 'test_{}_{}'.format(prefix, self._wrapped_connection.table_name)
        else:
            self._wrapped_connection.table_name = 'test_{}'.format(self._wrapped_connection.table_name)

    def __getattr__(self, attr):
        return getattr(self._wrapped_connection, attr)

    def update_item(self, *args, **kwargs):
        self._is_test_clean_required = True
        return self._wrapped_connection.update_item(*args, **kwargs)

    def put_item(self, *args, **kwargs):
        self._is_test_clean_required = True
        return self._wrapped_connection.put_item(*args, **kwargs)

    def batch_write_item(self, *args, **kwargs):
        self._is_test_clean_required = True
        return self._wrapped_connection.batch_write_item(*args, **kwargs)

    def post_test_clean(self, model_class):
        if self._is_test_clean_required:
            with model_class.batch_write() as batch:
                for item in model_class.scan():
                    batch.delete(item)
        self._is_test_clean_required = False
